# Benchmarks

All benchmarks were run on a Dell XPS 15 running Linux, using Amazon Corretto 11.0.9.

```
OpenJDK 64-Bit Server VM Corretto-11.0.9.11.1 (build 11.0.9+11-LTS, mixed mode)
```

## SLF4J Logger

As a baseline, here are the benchmarks for a straight SLF4J logger using Logback.

The average time is in comments.  For example, the baseline CPU time for a `logger.info("message")` is around 49 nanoseconds per operation, plus or minus 3 nanoseconds.

```java
public class SLF4JLoggerBenchmarks {
  private static final Logger logger =
      org.slf4j.LoggerFactory.getLogger(SLF4JLoggerBenchmarks.class);

  private static final Exception exception = new RuntimeException();

  @Benchmark
  public void info() {
    // SLF4JLoggerBenchmarks.info                    avgt    5   49.033 ± 3.101  ns/op
    logger.info("message");
  }
  
  @Benchmark
  public void isInfoEnabled() {
    // SLF4JLoggerBenchmarks.isInfoEnabled  avgt    5  0.949 ± 0.011  ns/op
   logger.isInfoEnabled();
  }

  @Benchmark
  public void infoWithArgument() {
    // SLF4JLoggerBenchmarks.infoWithArgument        avgt    5   49.749 ± 2.394  ns/op
    logger.info("message {}", "string");
  }

  @Benchmark
  public void infoWithArrayArgs() {
    // SLF4JLoggerBenchmarks.infoWithArrayArgs       avgt    5   51.215 ± 2.638  ns/op
    logger.info("message {} {} {} {}", "one", "two", "three", "four");
  }

  @Benchmark
  public void infoWithException() {
    // SLF4JLoggerBenchmarks.infoWithException       avgt    5  175.896 ± 0.807  ns/op
    logger.info("Message", exception);
  }
}
```

## Logstash Core Logger

The Logstash core logger wraps an SLF4J logger, backed by logback, and uses Logstash markers and structured arguments for fields.  This is not exposed directly, but is used internally.

Here, the `logback.xml` uses a no-op appender as follows:

```xml
<configuration>
    <statusListener class="ch.qos.logback.core.status.NopStatusListener"/>

    <appender name="NOP" class="ch.qos.logback.core.helpers.NOPAppender">
    </appender>

    <root level="DEBUG">
        <appender-ref ref="NOP" />
    </root>
</configuration>
```

The benchmarks are as follows:

```java
public class CoreLoggerBenchmarks {
  private static final CoreLogger logger = CoreLoggerFactory.getLogger();
  private static final Exception exception = new RuntimeException();
  private static final Field.Builder builder = Field.Builder.instance();

  @Benchmark
  public void info() {
    // CoreLoggerBenchmarks.info                     avgt    5   45.618 ± 0.965  ns/op
    logger.log(Level.INFO, "Message");
  }

  @Benchmark
  public void isEnabled() {
    // CoreLoggerBenchmarks.isEnabled       avgt    5  1.945 ± 0.049  ns/op
    logger.isEnabled(Level.INFO);
  }

  @Benchmark
  public void infoWithParameterizedString() {
    // CoreLoggerBenchmarks.infoWithParameterizedString  avgt    5   76.189 ± 2.388  ns/op
    logger.log(Level.INFO, "Message {}", fb -> fb.onlyString("foo", "bar"), builder);
  }

  @Benchmark
  public void infoWithException() {
    // CoreLoggerBenchmarks.infoWithException            avgt    5  174.242 ± 7.510  ns/op
    logger.log(Level.INFO, "Message", exception);
  }
}
```

The `Logger` uses a core logger internally, and has the same `logback.xml` configuration as the core logger.

```java
public class LoggerBenchmarks {
  private static final Logger<?> logger = LoggerFactory.getLogger();
  private static final Exception exception = new RuntimeException();

  @Benchmark
  public void info() {
    // LoggerBenchmarks.info                         avgt    5   47.464 ± 0.414  ns/op
    logger.info("Message");
  }

  @Benchmark
  public void isInfoEnabled(Blackhole blackhole) {
    // LoggerBenchmarks.isInfoEnabled       avgt    5  2.821 ± 0.063  ns/op
    blackhole.consume(logger.isInfoEnabled());
  }

  @Benchmark
  public void infoWithStringArg() {
    // LoggerBenchmarks.infoWithStringArg            avgt    5   75.672 ± 0.749  ns/op
    logger.info("Message", fb -> fb.onlyString("foo", "bar"));
  }

  @Benchmark
  public void infoWithContextString() {
    // LoggerBenchmarks.infoWithContextString        avgt    5  116.451 ± 6.000  ns/op
    logger.withFields(fb -> fb.onlyString("foo", "bar")).info("Message");
  }

  @Benchmark
  public void infoWithParameterizedString() {
    // LoggerBenchmarks.infoWithParameterizedString  avgt    5   75.732 ± 0.421  ns/op
    logger.info("Message {}", fb -> fb.onlyString("foo", "bar"));
  }

  @Benchmark
  public void infoWithException() {
    // LoggerBenchmarks.infoWithException            avgt    5  175.756 ± 1.843  ns/op
    logger.info("Message", exception);
  }

  @Benchmark
  public void infoWithNever() {
    // LoggerBenchmarks.infoWithNever                avgt    5   11.358 ± 0.165  ns/op
    logger.withCondition(Condition.never()).info("Message");
  }

  @Benchmark
  public void infoWithAlways() {
    // LoggerBenchmarks.infoWithAlways               avgt    5   57.670 ± 0.318  ns/op
    logger.withCondition(Condition.always()).info("Message");
  }

  @Benchmark
  public void infoWithFieldBuilder() {
    // LoggerBenchmarks.infoWithFieldBuilder         avgt    5   51.013 ± 0.436  ns/op
    logger.withFieldBuilder(Field.Builder.instance()).info("Message");
  }
}
```

## Log4J Implementation

The Log4J implementation using a `NullAppender`:

```java
public class Log4JBenchmarks {
    private static final Logger<?> logger = LoggerFactory.getLogger();

    private static final Exception exception = new RuntimeException();

    @Benchmark
    public void info() {
        // Log4JBenchmarks.info                         avgt    5  142.433 ± 27.654  ns/op
        logger.info("Message");
    }

    @Benchmark
    public void isInfoEnabled(Blackhole blackhole) {
        // Log4JBenchmarks.isInfoEnabled                avgt    5    5.556 ±  0.056  ns/op
        blackhole.consume(logger.isInfoEnabled());
    }

    @Benchmark
    public void infoWithStringArg() {
        // Log4JBenchmarks.infoWithStringArg            avgt    5  288.263 ±  0.753  ns/op
        logger.info("Message", fb -> fb.onlyString("foo", "bar"));
    }

    @Benchmark
    public void infoWithContextString() {
        // Log4JBenchmarks.infoWithContextString        avgt    5  247.300 ±  2.810  ns/op
        logger.withFields(fb -> fb.onlyString("foo", "bar")).info("Message");
    }

    @Benchmark
    public void infoWithParameterizedString() {
        // Log4JBenchmarks.infoWithParameterizedString  avgt    5  412.990 ± 27.346  ns/op
        logger.info("Message {}", fb -> fb.onlyString("foo", "bar"));
    }

    @Benchmark
    public void infoWithException() {
        // Log4JBenchmarks.infoWithException            avgt    5  238.892 ± 54.381  ns/op
        logger.info("Message", exception);
    }

    @Benchmark
    public void infoWithNever() {
        // Log4JBenchmarks.infoWithNever                avgt    5  176.454 ± 57.493  ns/op
        logger.withCondition(Condition.never()).info("Message");
    }

    @Benchmark
    public void infoWithAlways() {
        // Log4JBenchmarks.infoWithAlways               avgt    5  152.703 ± 52.607  ns/op
        logger.withCondition(Condition.always()).info("Message");
    }

    @Benchmark
    public void infoWithFieldBuilder() {
        // Log4JBenchmarks.infoWithFieldBuilder         avgt    5  156.345 ± 80.493  ns/op
        logger.withFieldBuilder(Field.Builder.instance()).info("Message");
    }
}
```

## Fluent Logger

The Fluent Logger uses the same configuration as above.  It is slower on average because object instantiation has some built-in costs, and takes 20 nanoseconds as a [baseline](https://shipilev.net/jvm/anatomy-quarks/6-new-object-stages/).

Using Logback:

```java
public class FluentBenchmarks {
  private static final FluentLogger<?> logger = FluentLoggerFactory.getLogger();
  private static final Exception exception = new RuntimeException();

  @Benchmark
  public void info() {
    // FluentBenchmarks.info                     avgt   25  103.478 ±  2.631  ns/op
    logger.atInfo().message("Message").log();
  }

  @Benchmark
  public void infoWithArgument() {
    // FluentBenchmarks.infoWithArgument         avgt   25  196.968 ±  4.884  ns/op
    logger.atInfo().message("Message {}").argument(fb -> fb.string("foo", "bar")).log();
  }

  @Benchmark
  public void infoWithArgAndException() {
    // FluentBenchmarks.infoWithArgAndException  avgt   25  331.651 ± 15.591  ns/op
    logger
        .atInfo()
        .message("Message {}")
        .argument(fb -> fb.string("foo", "bar"))
        .exception(exception)
        .log();
  }

  @Benchmark
  public void infoWithException() {
    // FluentBenchmarks.infoWithException        avgt   25  351.918 ±  7.124  ns/op
    logger.atInfo().message("Message").exception(exception).log();
  }
}
```

## Semantic Logger

The semantic logger configuration is the same as above.  There is only one kind of call you can make.

Using Logback:

```java
public class SemanticLoggerBenchmarks {

  private static final SemanticLogger<String> logger =
      SemanticLoggerFactory.getLogger(
          String.class, s -> "Message {}", s -> b -> b.onlyString("name", s));

  @Benchmark
  public void info() {
    // SemanticLoggerBenchmarks.info  avgt   25  97.199 ± 3.323  ns/op
    logger.info("string");
  }
}
```

## Scripting

Scripts can be read from file system or directly from memory.

```java
public class ScriptingBenchmarks {
    private static final Path path = Paths.get("src/jmh/tweakflow/condition.tf");

    private static final Path watchedDir = Paths.get("src/jmh/tweakflow");

    public static String buildScript() {
        StringBuilder b = new StringBuilder("");
        b.append("library echopraxia {");
        b.append("  function evaluate: (string level, dict fields) ->");
        b.append("    level == \"INFO\";");
        b.append("}");
        return b.toString();
    }

    private static final Condition fileCondition =
            ScriptCondition.create(false, path, Throwable::printStackTrace);

    private static final Condition stringCondition =
            ScriptCondition.create(false, buildScript(), Throwable::printStackTrace);

    private static final ScriptWatchService scriptWatchService = new ScriptWatchService(watchedDir);

    private static final ScriptHandle watchedScript =
            scriptWatchService.watchScript(
                    watchedDir.resolve("condition.tf"), Throwable::printStackTrace);

    private static final Condition watchedCondition = ScriptCondition.create(false, watchedScript);

    @Benchmark
    public void testFileConditionMatch(Blackhole blackhole) {
        // ScriptingBenchmarks.testFileConditionMatch     avgt    5  127.251 ± 0.816  ns/op
        blackhole.consume(fileCondition.test(Level.INFO, LogstashLoggingContext.empty()));
    }

    @Benchmark
    public void testStringConditionMatch(Blackhole blackhole) {
        // ScriptingBenchmarks.testStringConditionMatch   avgt    5  122.905 ± 3.440  ns/op
        blackhole.consume(stringCondition.test(Level.INFO, LogstashLoggingContext.empty()));
    }

    @Benchmark
    public void testFileConditionFail(Blackhole blackhole) {
        // ScriptingBenchmarks.testFileConditionFail      avgt    5  112.629 ± 2.951  ns/op
        blackhole.consume(fileCondition.test(Level.DEBUG, LogstashLoggingContext.empty()));
    }

    @Benchmark
    public void testStringConditionFail(Blackhole blackhole) {
        // ScriptingBenchmarks.testStringConditionFail    avgt    5  118.323 ± 4.296  ns/op
        blackhole.consume(stringCondition.test(Level.DEBUG, LogstashLoggingContext.empty()));
    }

    @Benchmark
    public void testWatchedConditionMatch(Blackhole blackhole) {
        // ScriptingBenchmarks.testWatchedConditionMatch  avgt    5  134.601 ± 2.325  ns/op
        blackhole.consume(watchedCondition.test(Level.INFO, LogstashLoggingContext.empty()));
    }

    @Benchmark
    public void testWatchedConditionFail(Blackhole blackhole) {
        // ScriptingBenchmarks.testWatchedConditionFail   avgt    5  125.652 ± 5.286  ns/op
        blackhole.consume(watchedCondition.test(Level.DEBUG, LogstashLoggingContext.empty()));
    }
}
```

where the `condition.tf` script is 

```
library echopraxia {
  function evaluate: (string level, dict fields) ->
     level == "INFO";
}
```